package items

import (
	"context"
	"fmt"
	"testing"
	"time"

	"git.perx.ru/perxis/perxis-go/pkg/schema"
	"git.perx.ru/perxis/perxis-go/pkg/schema/field"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
)

func TestItem_Set(t *testing.T) {
	item := &Item{}

	_ = item.Set("id", "id")
	assert.Equal(t, "id", item.ID)
	now := time.Now()

	_ = item.Set("created_at", now)
	assert.Equal(t, now, item.CreatedAt)

	_ = item.Set("a.b.c", 101)
	assert.Equal(t, map[string]any{"a": map[string]any{"b": map[string]any{"c": 101}}}, item.Data)

}

func TestItem_DeleteItemData(t *testing.T) {
	tests := []struct {
		name    string
		item    *Item
		field   string
		want    map[string]any
		wantErr assert.ErrorAssertionFunc
	}{
		{
			name:    "Simple",
			item:    &Item{Data: map[string]any{"a": "b", "c": "d"}},
			field:   "a",
			want:    map[string]any{"c": "d"},
			wantErr: assert.NoError,
		},
		{
			name:    "Item data is nil",
			item:    &Item{Data: nil},
			field:   "a",
			want:    nil,
			wantErr: assert.NoError,
		},
	}

	for _, tc := range tests {
		t.Run(tc.name, func(t *testing.T) {
			err := tc.item.Delete(tc.field)
			assert.NoError(t, err)
			if !tc.wantErr(t, err) {
				return
			}
			assert.Equal(t, tc.want, tc.item.Data)
		})
	}
}

func TestGetField(t *testing.T) {
	sch := schema.New(
		"a", field.String(),
		"obj", field.Object(
			"a", field.Number(field.NumberFormatFloat),
			"b", field.String(),
		),
		"arr", field.Array(field.Object("a", field.Time())),
	)
	sch.ClearState()

	tests := []struct {
		name    string
		field   string
		want    *field.Field
		wantErr assert.ErrorAssertionFunc
	}{
		{"Simple", "a", field.String(), assert.NoError},
		{"Incorrect field", "b", nil, assert.Error},
		{"Object", "obj", field.Object("a", field.Number(field.NumberFormatFloat), "b", field.String()), assert.NoError},
		{"Object path", "obj.a", field.Number(field.NumberFormatFloat), assert.NoError},
		{"Array", "arr", field.Array(field.Object("a", field.Time())), assert.NoError},
		{"Array path", "arr.a", field.Time(), assert.NoError},
		{"Array item", "arr.", field.Object("a", field.Time()), assert.NoError},
	}
	for _, tt := range tests {
		t.Run(tt.name, func(t *testing.T) {
			got, err := GetField(tt.field, sch)
			if !tt.wantErr(t, err, fmt.Sprintf("GetField(%v, sch)", tt.field)) {
				return
			}
			assert.Equalf(t, tt.want, got, "GetField(%v, sch)", tt.field)
		})
	}
}

func TestItem_Proto(t *testing.T) {
	w := time.Now().UTC()
	tests := []struct {
		name string
		item *Item
	}{
		{
			name: "All fields are filled",
			item: &Item{
				ID:           "id-1",
				SpaceID:      "space-1",
				EnvID:        "env-1",
				CollectionID: "coll-1",
				State:        StateDraft,
				CreatedRevAt: w.Add(time.Hour),
				CreatedBy:    "user-1",
				CreatedAt:    w,
				UpdatedAt:    w.Add(time.Hour),
				UpdatedBy:    "user-2",
				Data:         map[string]any{"a": "b", "c": "d", "x": nil},
				LocaleID:     "ru",
				Translations: map[string]map[string]interface{}{
					"ru": {"a": "B"},
					"en": nil,
				},
				TranslationsIDs:     []string{"ru", "en"},
				RevisionID:          "rev-1",
				RevisionDescription: "desc-1",
				Permissions:         PermissionsAllowAny,
				SearchScore:         100.0,
				Deleted:             false,
				Hidden:              true,
				Template:            false,
			},
		},
	}

	for _, tc := range tests {
		t.Run(tc.name, func(t *testing.T) {
			assert.Equal(t, tc.item, ItemFromProto(ItemToProto(tc.item)))
		})
	}
}

func TestItem_Encode_Decode(t *testing.T) {
	w := time.Now().UTC()
	tests := []struct {
		name string
		item *Item
	}{
		{
			name: "Data",
			item: &Item{
				ID:           "id-1",
				SpaceID:      "space-1",
				EnvID:        "env-1",
				CollectionID: "coll-1",
				State:        StateDraft,
				CreatedRevAt: w.Add(time.Hour),
				CreatedBy:    "user-1",
				CreatedAt:    w,
				UpdatedAt:    w.Add(time.Hour),
				UpdatedBy:    "user-2",
				Data: map[string]any{
					"a": "text-a",
					"b": 124.1,
					"c": map[string]interface{}{"x": "y"},
					"d": []interface{}{"k", "l", "m"},
				},
				RevisionID:          "rev-1",
				RevisionDescription: "desc-1",
				Permissions:         PermissionsAllowAny,
				SearchScore:         100.0,
				Deleted:             false,
				Hidden:              true,
				Template:            false,
			},
		},
		{
			name: "Data and Translations",
			item: &Item{
				ID:           "id-1",
				SpaceID:      "space-1",
				EnvID:        "env-1",
				CollectionID: "coll-1",
				State:        StateDraft,
				CreatedRevAt: w.Add(time.Hour),
				CreatedBy:    "user-1",
				CreatedAt:    w,
				UpdatedAt:    w.Add(time.Hour),
				UpdatedBy:    "user-2",
				Data: map[string]any{
					"a": "text-a",
					"b": 124.1,
					"c": map[string]interface{}{"x": "y"},
					"d": []interface{}{"k", "l", "m"},
				},
				LocaleID: "ru",
				Translations: map[string]map[string]interface{}{
					"ru": {"a": "ru-a"},
					"en": {"a": "en-a"},
				},
				TranslationsIDs:     []string{"ru", "en"},
				RevisionID:          "rev-1",
				RevisionDescription: "desc-1",
				Permissions:         PermissionsAllowAny,
				SearchScore:         100.0,
				Deleted:             false,
				Hidden:              false,
				Template:            false,
			},
		},
		{
			name: "Nil Translation",
			item: &Item{
				ID:           "id-1",
				SpaceID:      "space-1",
				EnvID:        "env-1",
				CollectionID: "coll-1",
				State:        StateDraft,
				CreatedRevAt: w.Add(time.Hour),
				CreatedBy:    "user-1",
				CreatedAt:    w,
				UpdatedAt:    w.Add(time.Hour),
				UpdatedBy:    "user-2",
				Data: map[string]any{
					"a": "text-a",
					"b": 124.1,
					"c": map[string]interface{}{"x": "y"},
					"d": []interface{}{"k", "l", "m"},
				},
				LocaleID: "ru",
				Translations: map[string]map[string]interface{}{
					"ru": {"a": "ru-a"},
					"en": nil,
				},
				TranslationsIDs:     []string{"ru", "en"},
				RevisionID:          "rev-1",
				RevisionDescription: "desc-1",
				Permissions:         PermissionsAllowAny,
				SearchScore:         100.0,
				Deleted:             false,
				Hidden:              false,
				Template:            false,
			},
		},
	}

	for _, tc := range tests {
		t.Run(tc.name, func(t *testing.T) {
			ctx := context.Background()
			sch := schema.New(
				"a", field.String(),
				"b", field.Number(field.NumberFormatFloat),
				"c", field.Object("x", field.String()),
				"d", field.Array(field.String()),
			)
			enc, err := tc.item.Encode(ctx, sch)
			require.NoError(t, err)
			dec, err := enc.Decode(ctx, sch)
			require.NoError(t, err)
			assert.Equal(t, tc.item, dec)
		})
	}
}
